import selectors
import socket
import traceback
import time

from ..core.packets_handler import PacketsHandler
from ..core.device import Device
from ..core.observer import dispatcher
from ..core.events import ApplicationEvents, DeviceEvents
from ..core.logger import log
from ..core.message.base import Handshake, Message
from .base import DeviceConnector


DEFAULT_HOST = "localhost"
DEFAULT_PORT = 9588
DEFAULT_HEADER_LEN = 4
DEFAULT_MAX_MSG_LEN = 4096



class RecvStatus:
    def __init__(self, device):
        self.device = device
        self.expected = 0
        self.binary_data = []


class SocketServerConnector(DeviceConnector):

    def __init__(self, config: dict = {}) -> None:
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self._selector = selectors.DefaultSelector()

        self._host = config.get("host", DEFAULT_HOST)
        self._port = config.get("port", DEFAULT_PORT)
        self._pck_max_len = config.get("pck_max_len", DEFAULT_MAX_MSG_LEN)

        self._run = False

        self._handler = PacketsHandler()

        dispatcher.add_observer(ApplicationEvents.ON_QUIT, self.on_exit)

    def on_exit(self):
        self._run = False

    def open(self) -> None:
        self._socket.bind((self._host, self._port))
        self._socket.listen()
        self._socket.setblocking(False)

        log(f"Socket Server Listenning on {self._host}:{self._port}")

        self._selector.register(self._socket, selectors.EVENT_READ, data=None)

    def close(self) -> None:
        log(f"Socket Server Closing")
        self._socket.close()
        self._selector.close()

    def new_device(self, socket: socket.socket) -> None:
        conn, addr = socket.accept()  # Should be ready to read

        device = Device()

        conn.setblocking(False)
        events = selectors.EVENT_READ | selectors.EVENT_WRITE
        self._selector.register(conn, events, data=RecvStatus(device))

        # JDE: the connected event is sent only whe, the device is ready 

    def on_device_msg(self, device: Device, socket: socket.socket, msg: Message) -> None:
        msg_str = msg.to_json()
        if len(msg_str) > 128:
            log(f"Socket Server Got message from {device.name}: {msg_str[:64]} ... {msg_str[-64:]}")
        else:
            log(f"Socket Server Got message from {device.name}: {msg_str}")

        dispatcher.notify(DeviceEvents.ON_MESSAGE, device, msg)

    def read_device(self, socket: socket.socket, status: RecvStatus) -> Message:
        log(f"Socket Server Got Msg")

        if status.expected == 0:
            length_data = []
            while len(length_data) < DEFAULT_HEADER_LEN:
                try:
                    length_data += socket.recv(DEFAULT_HEADER_LEN - len(length_data))
                except BlockingIOError:
                    return None
                except ConnectionResetError as err:
                    log(f"Socket Connector: Connection Reset Error Reading Device {err}", logger_level="ERROR")
                    raise err
                
                if len(length_data) == 0:
                    raise RuntimeError("Socket was closed")

            status.expected = int.from_bytes(length_data, "little")
            status.binary_data = []

        log(f"Socket waiting for {status.expected-len(status.binary_data)} bytes.", logger_level="INFO")
        while len(status.binary_data) < status.expected:
            try:
                status.binary_data += socket.recv(status.expected - len(status.binary_data))
            except BlockingIOError:
                return None
            except ConnectionResetError as err:
                log(f"Socket Connector: Connection Reset Error Reading Device {err}", logger_level="ERROR")
                raise err
           
            if len(status.binary_data) == 0:
                raise RuntimeError("Socket was closed")
        
        status.expected = 0

        return self._handler.push(bytearray(status.binary_data.copy()))

    def handle_connection(self, key, mask) -> None:
        sock = key.fileobj
        status: RecvStatus = key.data

        did_something = False
        try:
            if mask & selectors.EVENT_READ:
                did_something = True
                try:
                    msg = self.read_device(sock, status)
                    if msg:
                        if type(msg) == Handshake:
                            log("Got handshake from device: %s" % msg.agent_id)
                            status.device.set_name(msg.hid,msg.agent_id)

                        if type(msg) == Message:
                            self.on_device_msg(status.device, sock, msg)
                except RuntimeError as e:
                    self._selector.unregister(sock)
                    log(f"Closing case 1.1 {str(e)}", logger_level="ERROR")
                    sock.close()
                    dispatcher.notify(DeviceEvents.DISCONNECTED, status.device)
                except ConnectionResetError as e:
                    log(f"Closing case 1.2 {str(e)}", logger_level="ERROR")
                    dispatcher.notify(DeviceEvents.DISCONNECTED, status.device)

            if mask & selectors.EVENT_WRITE:
                if len(status.device.out_data):
                    did_something = True
                    data = status.device.out_data[0]
                    l = len(data)
                    log(f"Socket sending {l} bytes to device.")
                    remaining = 4
                    while (remaining):
                        remaining -= sock.send(l.to_bytes(4, "little"))
                    remaining = l
                    while (remaining):
                        remaining -= sock.send(data)
                    status.device.out_data = status.device.out_data[1:]

        except (ConnectionResetError, BrokenPipeError):
            self._selector.unregister(sock)
            log("Closing case 2", logger_level="ERROR")
            sock.close()
            dispatcher.notify(DeviceEvents.DISCONNECTED, status.device)
        except BlockingIOError:
            pass
        except Exception as e: 
            log(e, logger_level="ERROR")
            # raise e

        return did_something

    def run(self) -> None:
        try:
            self._run = True
            while self._run:
                ev = self._selector.select(timeout=1)
                if len(ev) == 0:
                    time.sleep(0.0001) # nothing was done so we give back then hand to the OS
                for key, mask in ev:
                    if key.data is None:
                        self.new_device(key.fileobj)
                    else:
                        if not self.handle_connection(key, mask):
                            time.sleep(0.0001) # nothing was done so we give back then hand to the OS
        except KeyboardInterrupt:
            self._run = False
        except Exception as e:
            log(e, logger_level="ERROR")
            traceback.print_exc()
        finally:
            log("Closing case 3")
            self.close()



# coding: utf-8

import os


def dir_exists(dir_path):
    return os.path.isdir(dir_path)

def file_exists(file_path):
    return os.path.exists(file_path)

def create_dir(dir_path):
    if not dir_exists(dir_path):
        os.makedirs(dir_path)

def get_sub_dir(dir_path):
    results = []
    if dir_exists(dir_path):
        for f in os.listdir(dir_path):
            if os.path.isdir(dir_path + "/" + f):
                results.append(f)
    return results

def get_files(dir_path, extension=None):
    results = []
    if dir_exists(dir_path):
        for f in os.listdir(dir_path):
            if os.path.isfile(dir_path + "/" + f) and (not extension or f.endswith("." + extension)):
                results.append(f)
    return results

def remove_file(file_path):
    if file_exists(file_path):
        os.remove(file_path)



import traceback
from serial import Serial
import threading

from ..core.packets_handler import PacketsHandler
from ..core.device import Device
from ..core.observer import dispatcher
from ..core.events import ApplicationEvents, DeviceEvents
from ..core.logger import log
from ..core.message.base import Handshake, Message
from .base import DeviceConnector


DEFAULT_PORT = "COM3"
MAX_PAYLOAD_SIZE = 4 * 1024





class RecvStatus:
    def __init__(self, device):
        self.device = device
        self.expected = 0
        self.binary_data = []


class SerialConnector(DeviceConnector):

    def __init__(self, config: dict = {}) -> None:
        self._port = config.get("port", DEFAULT_PORT)
        self._serial = None
        self._run = False

        self._handler = PacketsHandler()

        dispatcher.add_observer(ApplicationEvents.ON_QUIT, self.on_exit)

    def on_exit(self):
        self._run = False

    def open(self) -> None:
        self._serial = Serial(self._port)
        self._serial.baudrate = 115200;


    def close(self) -> None:
        log(f"Closing serial port")
        self._serial.close()

    def new_device(self) -> None:
        device = Device()
        dispatcher.notify(DeviceEvents.CONNECTED, device)
        x = threading.Thread(target=self._device_sender, args=(device,),daemon=True)
        x.start()

        return device

    def read_device(self) -> Message:
        length_array = self._serial.read(4)
        length =  int.from_bytes(length_array, "little")
        data = self._serial.read(length)

        try:
            msg = self._handler.push(data)
        except Exception as e:
            log(f"Error while handling packet: {e}")
            return None
        return msg

    def _device_sender(self,device):
        
        log("Serial Sender waiting for data...")
        while(self._run):
            if len(device.out_data) == 0:
                continue

            data = device.out_data[0]
            length =  len(data)
            if length > MAX_PAYLOAD_SIZE:
                log(f"Payload too big, max size is {MAX_PAYLOAD_SIZE} bytes")
                raise RuntimeError("Payload too big, max size is %d bytes" % MAX_PAYLOAD_SIZE)

            log(f"Sending {length} bytes to device")
            l1 = self._serial.write(length.to_bytes(4, "little"))
            self._serial.flush()
            l2 = self._serial.write(data)
            self._serial.flush()
            log(f"Sent {l1} header + {l2} payload to device")

            device.out_data = device.out_data[1:]
        log("Serial Sender ended")

    def run(self) -> None:
        log(f"Listening to serial port {self._port}")

        try:
            device = self.new_device()
            self._run = True
            while self._run:
                msg = self.read_device()
                if msg:
                    if type(msg) == Handshake:
                        log("Got handshake from device: %s" % msg.to_json())
                        device.set_name(msg.agent_id)

                    if type(msg) == Message:
                        log("Got message from device: %s" % msg.to_json())
                        dispatcher.notify(DeviceEvents.ON_MESSAGE, device, msg)


        except KeyboardInterrupt:
            self._run = False
        finally:
            log("Closing case 3")
            traceback.print_exc()
            self.close()



# coding: utf-8

import logging
import logging.config
import inspect


CONSOLE_LOGGER_NAME = "ConsoleLogger"
DEFAULT_LOGGER_NAME = CONSOLE_LOGGER_NAME


class LogsCustomConfig(object):
    default_logger: str


def init_logging(config: dict = None) -> None:
    if config:
        logging.config.dictConfig(config)
    LogsCustomConfig.default_logger = config.get("default_logger", DEFAULT_LOGGER_NAME)


def get_caller_name(idx: int = 2) -> str:
    stk = inspect.stack()
    
    if len(stk) < idx:
        frm = inspect.stack()[2]
        mod = inspect.getmodule(frm[0])
        return mod.__name__
    
    return __name__


# Method used to log
# Log engine is configured throught the config.yaml application file
# Users can add different logger to different outputs or different formats
# To log a msg just use log("msg")
# You can add parameters to configure where and how to log
# params:
#   - logger_name: Logger configured in config, for exemple : "ConsoleLogger"
#           default: "ConsoleLogger"
#   - logger_level: Logger Debug Level. Can be : "DEBUG", "INFO", "ERROR", "WARNING", "CRITICAL" or ints or lowercases
#           default: "info"
#   - *args and **kwargs forwarded to python logger engine

def log(msg: str, *args, **kwargs) -> None:
    logger_name = LogsCustomConfig.default_logger
    logger_level = "info"

    if "logger_name" in kwargs.keys():
        logger_name = kwargs["logger_name"]
        del kwargs["logger_name"]
    if "logger_level" in kwargs.keys():
        logger_level = kwargs["logger_level"]
        del kwargs["logger_level"]
        
    logger = logging.getLogger(logger_name)

    logger_func = logger.info
    if logger_level in [0, "DEBUG", "debug"]:
        logger_func = logger.debug
    elif logger_level in [1, "INFO", "info"]:
        logger_func = logger.info
    elif logger_level in [2, "WARNING", "warning"]:
        logger_func = logger.warning
    elif logger_level in [3, "ERROR", "error"]:
        logger_func = logger.error
    elif logger_level in [4, "CRITICAL", "critical"]:
        logger_func = logger.critical

    logger_func(msg, *args, **kwargs)



import asyncio
from websockets import WebSocketServerProtocol, serve

from ..logger import log
from ..observer import dispatcher
from ..events import WebsocketEvents
from ..websocket.client import WebsocketClient
from ..websocket.request import WebsocketRequest
from ..websocket.utils import ws_client_deserialize


DEFAULT_HOST="localhost"
DEFAULT_PORT=8765


class WebSocketConnector(object):
    def __init__(self, config: dict = {}) -> None:
        self._clients: set[WebsocketClient] = set()
        self._sockets: set[WebSocketServerProtocol] = set()
        self._host: str = config.get("host", DEFAULT_HOST)
        self._port: int = config.get("port", DEFAULT_PORT)

    async def on_message(self, socket: WebSocketServerProtocol) -> None:
        # We got a new client
        client = WebsocketClient(socket)
        self._clients.add(client)
        dispatcher.notify(WebsocketEvents.CONNECTED, client)
        log(f"WebSocket Server Got New Client {client.id}")

        # We'll loop asynchronously while client is connected
        # reading request it could send
        try:
            async for msg in socket:
                request = self.parse_request(msg)
                if request.is_valid():
                    dispatcher.notify(WebsocketEvents.ON_REQUEST, client=client, request=request)
        except Exception as err:
            log(f"Websocket error {err}", logger_level="ERROR")
            
        dispatcher.notify(WebsocketEvents.DISCONNECTED, client)
        try:
            self._clients.remove(client)
        except Exception as err:
            log(f"err removing Websocket {err}", logger_level="ERROR")
    
    async def run(self) -> None:
        # Serving server to host
        log(f"WebSocket Server Listenning on {self._host}:{self._port}")
        async with serve(self.on_message, self._host, self._port, max_size=None):
            await asyncio.Future()

    async def async_send_to_socket(self, socket: WebSocketServerProtocol, msg: str) -> None:
        await socket.send(msg)

    def send_to_socket(self, socket: WebSocketServerProtocol, msg: str) -> None:
        asyncio.create_task(self.async_send_to_socket(socket, msg))

    def parse_request(self, data: str) -> WebsocketRequest:
        try:
            # Loading json from recieved message
            # The mapping it to a request object
            # TODO: Maybe check that the request is valid here? 
            request_data = ws_client_deserialize(data)
            request = WebsocketRequest(**request_data)
            return request
        except Exception as e:
            log(f"Error parsing request data {str(e)}", logger_level="ERROR")



import asyncio
import atexit
import importlib
import yaml
import os
from inspect import isclass
from pathlib import Path
from yaml.loader import SafeLoader
import threading
from uuid import UUID

from .device import Device
from .logger import init_logging, log
from .manager import Manager
from .message.base import Message
from .websocket.utils import ws_client_serialize
from .events import DeviceEvents, WebsocketEvents
from .observer import dispatcher
from .websocket.connector import WebSocketConnector
from .websocket.client import WebsocketClient
from .websocket.request import ACT_GET_DEVICES_LIST, ACT_SEND_TO_DEVICE, ACT_SUB_TO_DEVICE, ACT_UNSUB_FROM_DEVUCE, \
            ACT_TRANSFERS_TO_WS_CLIENTS, EVT_ON_CLIENT_MSG, EVT_ON_DEVICE_MSG, WebsocketRequest
from .fsm import get_files

from ..device_connectors.base import DeviceConnector


CONFIG_DEFAULT_PATH: Path = Path(__file__).parent.parent / "config.yml"
CONNECTORS_PATH: Path = Path(__file__).parent.parent / "device_connectors"


class Application(object):
    def __init__(self, config_file_path: Path = CONFIG_DEFAULT_PATH) -> None:
        # Loading Configs
        self.config = {}
        self.load_config(config_file_path)
        init_logging(self.config.get("logs", None))

        # Init Device Manager
        Manager.Init()

        # Websocket server
        self.websocket_handler = WebSocketConnector(self.config.get("websocketserver", {}))

        # Device connectors
        self._connectors: list[DeviceConnector] = []
        self.load_connectors()

        # Adding event observers
        dispatcher.add_observer(WebsocketEvents.ON_REQUEST, self.handle_ws_client_request)
        dispatcher.add_observer(WebsocketEvents.CONNECTED, self.on_websocket_client_connected)
        dispatcher.add_observer(WebsocketEvents.DISCONNECTED, self.on_websocket_client_disconnected)
        dispatcher.add_observer(DeviceEvents.CONNECTED, self.on_device_connected)
        dispatcher.add_observer(DeviceEvents.DISCONNECTED, self.on_device_disconnected)
        dispatcher.add_observer(DeviceEvents.ON_MESSAGE, self.on_device_message)

    def load_config(self, config_file: Path) -> None:
        # Loading yaml config file
        try:
            with open(config_file) as f:
                self.config = yaml.load(f, Loader=SafeLoader)
        except Exception as e:
            raise RuntimeError(str(e))

    def load_connectors(self) -> None:
        # Loading active connectors from config
        config_connectors = self.config.get("connectors", {})
        active_connector_names = [k for k in config_connectors.keys() if config_connectors[k].get("active", False)]

        # Getting all python files in our connector directory path
        for connector_name in get_files(str(CONNECTORS_PATH), "py"):
            if connector_name not in ["__init__.py", "base.py", "__pycache__"]:
                # Loading the python module file
                connector_module_name = connector_name.replace(".py", "")
                # log(f"Application Loading Connector from module {connector_module_name}")
                mo = importlib.import_module(f".device_connectors.{connector_module_name}", package="bondzai.gateway")

                # Getting connectors classes from loaded module
                connector_cls = None
                for attr in dir(mo):
                    if attr != "DeviceConnector" and attr in active_connector_names:
                        log(f"Application Loading Connector {attr}")
                        connector_cls = getattr(mo, attr)
                        # If our object is a class and a sub class of DeviceConnector
                        # Then we add it to our connectors list
                        if isclass(connector_cls) and issubclass(connector_cls, DeviceConnector):
                            connector = connector_cls(config_connectors[attr].get("config", {}))
                            self._connectors.append(connector)

    def run(self) -> None:
        # Run open method for each active connectors
        for instance in self._connectors:
            instance.open()
        
        # TODO : Find Windows solution
        def clean_exit():
            log("Exiting application")
            # Cleanning active connectors
            for c in self._connectors:
                c.close()
        atexit.register(clean_exit)

        # Method for WebSocketServer Thread
        def run_websockets_server():
            asyncio.run(self.websocket_handler.run())

        threads = []

        # Creating Thread for WebSocketServer
        t = threading.Thread(target=run_websockets_server, daemon=True)
        threads.append(t)
        # Launching thread
        t.start()

        for c in self._connectors:
            # Creating thread for active connector
            x = threading.Thread(target=c.run, daemon=True)
            threads.append(x)
            # Launching thread
            x.start()
        
        try:
            for x in threads:
               # Waiting threads to finish
                x.join()
        except KeyboardInterrupt:
            pass

    # Callback for WebSocketServer event ON_REQUEST
    def handle_ws_client_request(self, client: WebsocketClient, request: WebsocketRequest):
        log(f"Application Got Client ({client.id}) Request : {request.to_string()}")

        # Handle client request. Could be:
        #   - Register to device events
        #   - Unregister from device events
        #   - Get Device List
        #   - Send to device

        if request.action == ACT_SUB_TO_DEVICE:
            Manager.SubToDeviceEvents(request.device_name, client)
        elif request.action == ACT_UNSUB_FROM_DEVUCE:
            Manager.UnsubFromDeviceEvents(request.device_name, client)
        elif request.action == ACT_GET_DEVICES_LIST:
            self.send_device_list_to_clients(client)
        elif request.action == ACT_TRANSFERS_TO_WS_CLIENTS:
            self.transfers_msg_to_clients(request.message, client.id, request.to_client_id)
        elif request.action == ACT_SEND_TO_DEVICE:
            device = Manager.GetDevice(request.device_name)
            if device:
                msg = Message.from_json(request.message)

                # Force the session id to be the one generated at handshake
                msg.header.session = device.session.bytes

                device.send(msg.to_msgpack())

    def send_device_list_to_clients(self, client: WebsocketClient = None) -> None:
        # if a client is specified, we send the list only to him, else to all clients
        clients = [client] if client is not None else Manager.GetClients()
        device_list = [d.name for d in Manager.GetDevices()]
        msg = ws_client_serialize({
            "action": ACT_GET_DEVICES_LIST,
            "devices": device_list
        })

        for client in clients:
            client.send(msg)

    def transfers_msg_to_clients(self, msg: str, from_client_id: UUID, to_client: str = None) -> None:
        for client in Manager.GetClients():
            if client.id != from_client_id and (not to_client or str(to_client) == str(client.id)):
                log(f"Sending to {client.id}")
                client.send(ws_client_serialize({
                    "from_client_id": str(from_client_id),
                    "action": EVT_ON_CLIENT_MSG,
                    "message": msg
                }))

    # Callback for Device event CONNECTED
    def on_device_connected(self, device: Device) -> None:
        log(f"New device connected {device.name}")
        Manager.RegisterDevice(device)
        self.send_device_list_to_clients()

    # Callback for Device event DISCONNECTED
    def on_device_disconnected(self, device: Device) -> None:
        log(f"Device disconnected {device.name}")
        Manager.UnregisterDevice(device)
        self.send_device_list_to_clients()

    # Callback for Device Event ON_MESSAGE
    def on_device_message(self, device: Device, msg: Message) -> None:
        # TODO : Handle msg       
        # TMP : For now send msg to every subscribed websocket clients
        log(f"Application On Device Message {device.name}")
        for client in Manager.GetDeviceEventSubs(device.name):
            log(f"Application Forward")
            client.send(ws_client_serialize({ 
                "action": EVT_ON_DEVICE_MSG,
                "device_name": device.name,
                "message": msg.to_dict()
            }))

    # Callback when a websocket client is connected
    def on_websocket_client_connected(self, client: WebsocketClient) -> None:
        Manager.RegisterClient(client)

    # Callback when a websocket client is disconnected
    def on_websocket_client_disconnected(self, client: WebsocketClient) -> None:
        log(f"Websocket Client {client.id} disconnected")
        Manager.UnregisterClient(client)
        Manager.UnsubClient(client)



# Filename: packets_handler.py
# the class that handles packets coming from the device

from .message.base import Handshake, Message
from .device import TAG_AGENT_ID

class PacketsHandler(object):
    """
    This class is here to handle the packets (before handling the message)
    It allows to check packet header and to collatione the packet if message is not
    complete
    """
    def __init__(self):
        self._buffer = bytearray([])
        self._pid = -1

    def handle_handshake_message(self, header, payload ):        
        """
        This function is here to handle the applicative messages
        """
        agent_id = None
        pid = header[1]

        #look for AGENT_ID tag
        i = 0
        while i<len(payload):
            tag = payload[i]
            length = payload[i+1]
            value = payload[i+2:i+2+length]
            i += 2+length

            if tag == TAG_AGENT_ID:
                agent_id = value
                break
        
        if not agent_id:
            raise RuntimeError("No agent id in handshake message")
        
        return Handshake(pid,agent_id)



    def handle_app_message(self, header, payload ):        
        """
        This function is here to handle the applicative messages
        """

        islast = (header[0]&1)
        pid = header[1]

        self._pid = pid
        self._buffer += payload

        if not islast:
            return None

        self._pid = -1
        message = self._buffer
        self._buffer = bytearray([])

        return Message.from_msgpack(message)

         
    def push(self, packet):
        if not packet:
            raise RuntimeError("Connection error, empty packet")

        header = packet[:2]
        payload = packet[2:]

        pid = header[1]
        #check version
        version = ((header[0]>>4)&0xF)
        issecure = (header[0]&8)
        msgtyp = ((header[0]&6)>>1)

        if (version != 1):
            raise RuntimeError("Bad protocol version %d, only supporting version 1" % version)

        if (self._pid != -1 and pid != self._pid):
            raise RuntimeError("This version of gateway doesn't support interleaved messages")

        if issecure:
            raise RuntimeError("This version of gateway doesn't support secure messages")

        if (msgtyp == 0):
            return self.handle_handshake_message(header, payload)


        if (msgtyp == 1):
            return self.handle_app_message(header, payload)
       

        raise RuntimeError("Bad msg type %d, only supporting application message (1)" % msgtyp)



from dataclasses import dataclass

from ..logger import log
from .payload import MessagePayload
from .utils import B642B, unpack_buffer_to_list, B2B64, get_step_struct
from .enums import EventOperationID

@dataclass
class EventHeader:
    mod: int 
    appid: int 
    timestamp: int
    payload_size: int

    def to_dict(self) -> str:
        return {
            "mod": self.mod,
            "appid": self.appid,
            "timestamp": self.timestamp,
            "payload_size": self.payload_size
        }

    def to_array(self) -> list:
        return [
            self.mod, 
            self.appid,
            self.timestamp, 
            self.payload_size
        ]


@dataclass
class EventDataIn:
    datasource: int
    datatype: str
    buffer: bytes

    def to_dict(self) -> str:
        return {
            "datasource": self.datasource,
            "datatype": self.datatype,
            "buffer": unpack_buffer_to_list(self.buffer, self.datatype[0] == "<", 4, self.datatype[1])
        }

    def to_array(self) -> list:
        return [
            self.datasource,
            self.datatype,
            self.buffer
        ]


@dataclass
class EventException:
    error: int
    msg: str

    def to_dict(self) -> str:
        return {
            "error": self.error,
            "msg": self.msg
        }

    def to_array(self) -> list:
        return [
            self.error,
            self.msg
        ]


@dataclass
class EventCMDStatus:
    mod: int
    op: int
    status: int

    def to_dict(self) -> str:
        return {
            "mod": self.mod,
            "op": self.op,
            "status": self.status
        }

    def to_array(self) -> list:
        return [
            self.mod,
            self.op,
            self.status
        ]


@dataclass
class EventLog:
    level: int
    msg: str

    def to_dict(self) -> str:
        return {
            "level": self.level,
            "msg": self.msg
        }

    def to_array(self) -> list:
        return [
            self.level,
            self.msg
        ]


@dataclass
class EventAslResult:
    label_type: int
    key_id: int
    confidence: float
    step: int
    result: list

    def to_dict(self) -> str:

        step_fields = get_step_struct(self.step)

        return {
            "label_type": self.label_type, # keep it for compatibility, it's equal to output
            "confidence": self.confidence,
            "source_id": step_fields.get("source_id",0),
            "process_id": step_fields.get("process_id",0),
            "ai_mode": step_fields.get("ai_mode",0),
            "record_mode": step_fields.get("record_mode",0),
            "output": step_fields.get("output",0),
            "result": self.result,
            "key_id": self.key_id            
        }

    def to_array(self) -> list:
        return [
            self.label_type,
            self.key_id,        
            self.confidence,
            self.step,
            self.result
        ]


@dataclass
class EventInferResult:
    ai_type:int
    label_type: int
    key_id: int
    step: int
    result: list

    def to_dict(self) -> str:
        step_fields = get_step_struct(self.step)

        return {
            "ai_type": self.ai_type,
            "label_type": self.label_type,
            "source_id": step_fields.get("source_id",0),
            "process_id": step_fields.get("process_id",0),
            "ai_mode": step_fields.get("ai_mode",0),
            "record_mode": step_fields.get("record_mode",0),
            "output": step_fields.get("output",0),
            "result": self.result,
            "key_id": self.key_id            
        }

    def to_array(self) -> list:
        return [
            self.ai_type,
            self.label_type,
            self.key_id,
            self.step,
            self.result
        ]


@dataclass
class EventSetMode:
    ai_mode: int
    ai_type: int
    data: list

    def to_dict(self) -> str:
        return {
            "ai_mode": self.ai_mode,
            "ai_type": self.ai_type,
            "data":self.data
        }
    
    def to_array(self) -> list:
        return [
            self.ai_mode,
            self.ai_type,
            self.data
        ]


@dataclass
class EventCorrection:
    position: int
    source_id: int
    ai_type: int
    data: list  # classification.label_type & classification.label OR regression

    def to_dict(self) -> str:
        return {
            "position": self.position,
            "source_id": self.source_id,
            "ai_type": self.ai_type,
            "data": self.data
        }

    def to_array(self) -> list:
        return [
            self.position,
            self.source_id,
            self.ai_type,
            self.data
        ]

@dataclass
class EventTrigger:
    trigger_type: int
    trigger_value: int
    key_id: int

    def to_dict(self) -> str:
        return {
            "trigger_type": self.trigger_type,
            "trigger_value": self.trigger_value,
            "key_id": self.key_id,
        }
    
    def to_array(self) -> list:
        return [
            self.trigger_type,
            self.trigger_value,
            self.key_id
        ]
 
@dataclass
class EventSetRecordMode:
    record_mode: int
    source_id: int

    def to_dict(self) -> str:
        return {
            "record_mode": self.record_mode,
            "source_id": self.source_id,
        }
    
    def to_array(self) -> list:
        return [
            self.record_mode,
            self.source_id
        ]
    
@dataclass
class EventLaunchTrain:

    def to_dict(self) -> str:
        return { }
    
    def to_array(self) -> list:
        return [ ]

@dataclass
class EventKill:

    def to_dict(self) -> str:
        return { }
    
    def to_array(self) -> list:
        return [ ]

@dataclass
class EventCustom:
    data: bytearray

    def __init__(self, data) -> None:
        if(type(data) == str):
            data = B642B(data)
        self.data = bytearray(data)

    def to_dict(self) -> str:
        return { "data": B2B64(self.data) }

    def to_array(self) -> list:
        return [ self.data ]
    

@dataclass
class EventProcessAck:
    evt_id: int
    process_state: int
    step: int
    app_id: int
    meta: bytes

    def to_dict(self) -> str:
        step_fields = get_step_struct(self.step)
        return {
            "evt_id": self.evt_id,
            "process_state": self.process_state,
            "source_id": step_fields.get("source_id",0),
            "process_id": step_fields.get("process_id",0),
            "ai_mode": step_fields.get("ai_mode",0),
            "record_mode": step_fields.get("record_mode",0),
            "output": step_fields.get("output",0),
            "app_id": self.app_id,
            "meta": self.meta
        }

    def to_array(self) -> list:

        return [
            self.evt_id,
            self.process_state,
            self.step,
            self.app_id,
            self.meta
        ]


@dataclass
class EventDataResult:
    evt_id: int
    process_state: int
    step: int
    app_id: int
    datatype: str
    result: bytes

    def to_dict(self) -> str:
        listofresult = [unpack_buffer_to_list(b, self.datatype[0] == "<", 4, self.datatype[1]) for b in self.result]
        result = [item for onelist in listofresult for item in onelist]
        step_fields = get_step_struct(self.step)

        return {
            "evt_id": self.evt_id,
            "process_state": self.process_state,
            "source_id": step_fields.get("source_id",0),
            "process_id": step_fields.get("process_id",0),
            "ai_mode": step_fields.get("ai_mode",0),
            "record_mode": step_fields.get("record_mode",0),
            "output": step_fields.get("output",0),
            "app_id": self.app_id,
            "datatype": self.datatype,
            "result": result
        }

    def to_array(self) -> list:
        return [
            self.evt_id,
            self.process_state,
            self.step,
            self.app_id,
            self.datatype,
            self.result
        ]
    

@dataclass
class EventTrainResult:
    evt_id: int
    process_state: int
    step: int
    app_id: int
    nb_models: int

    def to_dict(self) -> str:
        step_fields = get_step_struct(self.step)

        return {
            "evt_id": self.evt_id,
            "process_state": self.process_state,
            "source_id": step_fields.get("source_id",0),
            "process_id": step_fields.get("process_id",0),
            "ai_mode": step_fields.get("ai_mode",0),
            "record_mode": step_fields.get("record_mode",0),
            "output": step_fields.get("output",0),
            "app_id": self.app_id,
            "nb_models": self.nb_models
        }

    def to_array(self) -> list:
        return [
            self.evt_id,
            self.process_state,
            self.step,
            self.app_id,
            self.nb_models
        ]    

OP_TO_EVENT_TYPES = {
    EventOperationID.EVT_EXT_EXCEPTION.value:     EventException,
    EventOperationID.EVT_EXT_DATA_IN.value:       EventDataIn,
    EventOperationID.EVT_EXT_CMD_STATUS.value:    EventCMDStatus,
    EventOperationID.EVT_EXT_LOG.value:           EventLog,
    EventOperationID.EVT_EXT_ASL_RESULT.value:    EventAslResult, 
    EventOperationID.EVT_EXT_VM_RESULT.value:     EventInferResult, 
    EventOperationID.EVT_EXT_SET_MODE.value:      EventSetMode,
    EventOperationID.EVT_EXT_CORRECTION.value:    EventCorrection,
    EventOperationID.EVT_EXT_TRIGGER.value:       EventTrigger,
    EventOperationID.EVT_EXT_SET_RECORD_MODE.value: EventSetRecordMode,
    EventOperationID.EVT_EXT_LAUNCH_TRAIN.value:  EventLaunchTrain,
    EventOperationID.EVT_EXT_KILL.value:          EventKill,
    EventOperationID.EVT_EXT_CUSTOM_1.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_2.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_3.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_4.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_5.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_6.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_7.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_8.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_9.value:      EventCustom,
    EventOperationID.EVT_EXT_CUSTOM_10.value:     EventCustom,
    EventOperationID.EVT_EXT_PROCESS_ACK.value:   EventProcessAck,
    EventOperationID.EVT_EXT_DATA_TRANSFORM_RESULT.value:   EventDataResult,
    EventOperationID.EVT_EXT_DATA_ACQ_RESULT.value:   EventDataResult,
    EventOperationID.EVT_EXT_ENROLL_RESULT.value: EventDataResult,
    EventOperationID.EVT_EXT_DATASET_PREPARATION_RESULT.value: EventTrainResult,
    EventOperationID.EVT_EXT_TRAINING_RESULT.value: EventTrainResult
}

def GetEventClass(event_type_id: int):
    return OP_TO_EVENT_TYPES.get(event_type_id, None)

class EventMessagePayload(MessagePayload):
    @classmethod
    def from_array(cls, operation: int, raw_data: list) -> "MessagePayload":
        EventClass = GetEventClass(operation)
        if EventClass is None:
            log(f"Operation {operation} unknown", logger_level="ERROR")
            return

        payload = cls()
        payload.header = EventHeader(*raw_data[:4])
        payload.event = EventClass(*raw_data[4:])

        return payload

    @classmethod
    def from_json(cls, operation: int, raw_data: list) -> "MessagePayload":
        EventClass = GetEventClass(operation)
        if EventClass is None:
            log(f"Operation {operation} unknown", logger_level="ERROR")
            return
            
        payload = cls()
        payload.header = EventHeader(**raw_data["header"])
        payload.event = EventClass(**raw_data["data"])

        return payload

    def to_dict(self) -> dict:
        return {
            "header": self.header.to_dict(),
            "data": self.event.to_dict()
        }



class MessagePayload(object):
    @classmethod
    def from_array(cls, operation: int, raw_data: list) -> "MessagePayload":
        raise NotImplemented

    @classmethod
    def from_json(cls, operation: int, raw_data: list) -> "MessagePayload":
        raise NotImplemented

    def to_dict(self) -> dict:
        raise NotImplemented



import msgpack
import sys, traceback
from enum import Enum
from io import BytesIO

from .utils import B2B64
from .enums import DBMTable
from ..logger import log



class DBM_ATT_VM(Enum):
    DESC                = 0x0
    PREPROC_GRAPH_TRAIN = 0x1
    PREPROC_GRAPH_INFER = 0x2
    PREPROC_GRAPH_OTHER = 0x3
    DEEPLOMATH          = 0x4
    POSTPROC            = 0x5


class DBM_ATT_DAT(Enum):
    META                = 0x0
    DATA                = 0x1


class DbmExportVm(object):
    def __init__(self, raw_data):
        self.attributes = {}

        for att in raw_data:
            if att == DBM_ATT_VM.DESC.value:
                i=0
                self.attributes["desc"] = {}
                self.attributes["desc"]["description"] = raw_data[att][i]
                i+=1
                if len(raw_data[att][i]):
                    self.attributes["desc"]["axis"] = raw_data[att][i]
                    i+=1
                else:
                    self.attributes["desc"]["axis"] = raw_data[att][i+1]
                    i+=2
                if len(raw_data[att][i]):
                    self.attributes["desc"]["split"] = raw_data[att][i]
                    i+=1
                else:
                    self.attributes["desc"]["split"] = raw_data[att][i+1]
                    i+=2
                self.attributes["desc"]["maxsplit"] = raw_data[att][i]
                i+=1
                self.attributes["desc"]["reduction"] = raw_data[att][i]
                i+=1
                self.attributes["desc"]["filter"] = {
                    "max_row": raw_data[att][i],
                    "mode": raw_data[att][i+1],
                    "filters": []
                } 
                i+=2

                for el in raw_data[att][i]:
                    filt =  {
                        "row" : el[0] if len(el[0]) else el[1],
                        "labels": [
                            lbl for lbl in (el[1] if len(el[0]) else el[2]) 
                        ]                        
                    }
                    self.attributes["desc"]["filter"]["filters"].append(filt)            
            elif att in [DBM_ATT_VM.PREPROC_GRAPH_TRAIN.value, DBM_ATT_VM.PREPROC_GRAPH_INFER.value, DBM_ATT_VM.PREPROC_GRAPH_OTHER.value]:
                graph = {
                    "memneed": {
                        "outlen": raw_data[att][0],
                        "templen": raw_data[att][1],
                    },
                    "sizeout": raw_data[att][2],
                }
                
                nodes = []                
                for n in raw_data[att][3]:
                    nodes.append({
                        "op_id": n[0],
                        "memneed": {
                            "outlen": n[1],
                            "templen": n[2]
                        },
                        "mem_policy": n[3],
                        "next": [i for i in n[4]],
                        "param": {
                            "typeid": n[5],
                            "parameters": B2B64(b"".join(n[6]))
                        }
                    })

                    k = "preproc_other"
                    if att == DBM_ATT_VM.PREPROC_GRAPH_TRAIN.value:
                        k = "preproc_train"
                    elif att==DBM_ATT_VM.PREPROC_GRAPH_INFER.value:
                        k = "preproc_infer" 

                    graph["nodes"] = nodes
                    self.attributes[k] = graph
            elif att == DBM_ATT_VM.DEEPLOMATH.value :
                self.attributes["deeplomath"] = {
                    "mode": raw_data[att][0],
                    "rejection": raw_data[att][1],
                    "dim_in": raw_data[att][2],
                    "dim_out": raw_data[att][3],
                    "max_layers": raw_data[att][4]
                }
            elif att == DBM_ATT_VM.POSTPROC.value :
                self.attributes["postproc"] = {
                    "mode": raw_data[att][0],
                    "rejection": raw_data[att][1],
                    "op": raw_data[att][2],
                    "ring_depth": raw_data[att][3],
                    "ring_width": raw_data[att][4],
                    "reserved": raw_data[att][5]
                }

    def to_dict(self):
        return self.attributes


class DbmExportLbl(object):
    pass


class DbmExportDat(object):
    def __init__(self, raw_data):
        self.attributes={}
        for att in raw_data:
            if att == DBM_ATT_DAT.META.value:
                self.attributes["label"] = {
                    "type": raw_data[att][0],
                    "qi": raw_data[att][1]
                } 

                if self.attributes["label"]["type"] == 0:                
                    self.attributes["label"]["labels"] = [{
                        "output_id": l[0],
                        "source_id": l[1],
                        "label": l[2]
                    } for l in raw_data[att][2]]
                else:
                    self.attributes["label"]["values"] = raw_data[att][2]                
            elif att == DBM_ATT_DAT.DATA.value:

                raw_data_l = raw_data[att][0]
                raw_data_bytes = raw_data[att][1]
                self.attributes["record"] = B2B64(b"".join(raw_data_bytes))\

    def to_dict(self):
        return self.attributes


class DbmExportKey(object):
    pass


class DbmExportCtx(object):
    pass


class DbmExportSig(object):
    pass


class DbmExportGen(object):
    pass


DBM_TYPE_TO_CLASS = {
    DBMTable.DBM_VM  : DbmExportVm, 
    DBMTable.DBM_LBL : DbmExportLbl,
    DBMTable.DBM_DAT : DbmExportDat,
    DBMTable.DBM_KEY : DbmExportKey,
    DBMTable.DBM_CTX : DbmExportCtx,
    DBMTable.DBM_SIG : DbmExportSig,
    DBMTable.DBM_GEN : DbmExportGen
}


class DbmProcesser(object):
    @classmethod
    def from_msgpack(cls, data: bytearray) -> "DbmProcesser":
        try:

            stream = BytesIO(data)
            unpacker = msgpack.Unpacker(stream, raw=False, strict_map_key=False)

            table_type = unpacker.unpack()

            nb_attributes = unpacker.read_map_header()
            
            att_map = {}
            for _ in range(nb_attributes):
                k = unpacker.unpack()
                v = unpacker.unpack()
                att_map[k] = v

            return DBM_TYPE_TO_CLASS[DBMTable(table_type)](att_map)
        except Exception as e:
            log(f"Error parsing msgpck in DbmProcesser : {str(e)}", logger_level="ERROR")
            traceback.print_exc(file=sys.stdout)
            return None



from enum import Enum
from dataclasses import dataclass
from io import BytesIO
import msgpack

from ..logger import log

from .event import EventMessagePayload
from .command import CommandMessagePayload
from .response import ResponseMessagePayload
from .payload import MessagePayload
from .utils import B2B64, msg_deserialize, msg_serialize, pack_list_to_buffer, B642B


class MSG_TYP(Enum):
    COMMAND     = 0
    RESPONSE    = 1
    ERR         = 2
    EVENT       = 3


@dataclass
class MsgHeader:
    session: bytes
    id: int
    op: int
    typ: int
    mod: int

    def to_dict(self) -> str:
        return {
            "session": B2B64(self.session),
            "id": self.id,
            "op": self.op,
            "typ": self.typ,
            "mod": self.mod
        }

    def to_array(self) -> list:

        #make sure session is 32 bytes, else pad
        if len(self.session) < 32:
            self.session = self.session + bytearray(32-len(self.session))

        return [
            self.session, 
            self.id, 
            self.op, 
            self.typ, 
            self.mod
        ]


class Message(object):
    def __init__(self, raw_data: list = None) -> None:
        self.header: MsgHeader = MsgHeader(*raw_data[:5]) if raw_data is not None else None
        self.payloads: list[MessagePayload] = []

        if raw_data is not None:
            payload = None
            if self.header.typ == MSG_TYP.EVENT.value:
                payload = EventMessagePayload.from_array(
                    self.header.op, 
                    raw_data[5:]
                )
            elif self.header.typ == MSG_TYP.COMMAND.value:
                payload = CommandMessagePayload.from_array(
                    self.header.mod,
                    self.header.op, 
                    raw_data[5:]
                )
            elif self.header.typ == MSG_TYP.RESPONSE.value:
                payload = ResponseMessagePayload.from_array(
                    self.header.mod,
                    self.header.op, 
                    raw_data[5:]
                )
            else:
                log(f"Unknown type {str(self.header.typ)}", logger_level="ERROR")
                
            if payload is not None:
                self.payloads.append(payload)

    def to_dict(self) -> dict:
        res = {
            "header": self.header.to_dict(),
            "payloads": [p.to_dict() for p in self.payloads]
        }
        return res

    def to_json(self) -> str:
        return msg_serialize(self.to_dict())

    @classmethod
    def from_json(cls, json_data: str) -> "Message":
        try:
            data = msg_deserialize(json_data)
        except Exception as e:
            log(f"Error parsing json in {str(e)}", logger_level="ERROR")
            return None

        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data: dict) -> "Message":
        if "header" not in data:
            log("Missing Header in data", logger_level="ERROR")
            return None
        
        if "payloads" not in data:
            log("Missing payloads in data", logger_level="ERROR")
            return None
            
        if isinstance(data["header"]["session"], str):
            data["header"]["session"] = B642B(data["header"]["session"])

        m = Message()
        m.header = MsgHeader(**data["header"])

        for payload in data["payloads"]:
            #FIXME: (JDE) I think this should be done in the to_array function of the class
            #if "buffer" in payload["data"] and type(payload["data"]["buffer"][0]) == (float):                
            if "buffer" in payload["data"] and "datatype" in payload["data"] and payload["data"]["datatype"][-1] == "f":
                payload["data"]["buffer"] = pack_list_to_buffer(payload["data"]["buffer"], True, "f")

            if m.header.typ == MSG_TYP.COMMAND.value:
                p = CommandMessagePayload.from_json(
                    data["header"]["mod"],
                    data["header"]["op"], 
                    payload                    
                )
            elif m.header.typ == MSG_TYP.EVENT.value:
                p = EventMessagePayload.from_json(
                    data["header"]["op"], 
                    payload
                )
            else:
                log(f"This message type is not supported : {str(m.header)}", logger_level="ERROR")

            m.payloads.append(p)

        return m

    @classmethod
    def from_msgpack(cls, data: bytearray) -> "Message":
        try:
            raw_data = []

            unpacker = msgpack.Unpacker(BytesIO(data), raw=False)
            for unpacked in unpacker:
                raw_data.append(unpacked)

            return cls(raw_data)
        except Exception as e:
            log(f"Error parsing msgpck in {str(e)}", logger_level="ERROR")
            return None

    def to_msgpack(self) -> bytes:
        raw_data = self.header.to_array()

        for payload in self.payloads:
            if "header" in payload.__dict__:
                raw_data.extend(payload.header.to_array())
            if "event" in payload.__dict__:
                raw_data.extend(payload.event.to_array())
            if "command" in payload.__dict__:
                raw_data.extend(payload.command.to_array())

        packer = msgpack.Packer()
        buffer = BytesIO()
        for i in range(len(raw_data)):
            data = raw_data[i]
            buffer.write(packer.pack(data))
        buffer.seek(0)

        return buffer.read()


class Handshake(object):

    def __init__(self, hid: int, agent_id: bytes) -> None:
        self.hid = hid
        self.agent_id = agent_id




from enum import Enum


class WebsocketEvents(Enum):
    ON_REQUEST = "WEBSOCKET-ON-REQUEST"
    CONNECTED = "WEBSOCKET-CLIENT-CONNECTED"
    DISCONNECTED = "WEBSOCKET-CLIENT-DISCONNECTED"

    def __repr__(self) -> str:
        return self.value


class DeviceEvents(Enum):
    CONNECTED = "DEVICE_CONNECTED"
    DISCONNECTED = "DEVICE_DISCONNECTED"

    ON_MESSAGE = "ON_DEVICE_MESSAGE"

    def __repr__(self) -> str:
        return self.value


class ApplicationEvents(Enum):
    ON_QUIT = "APPLICATION-ON-QUIT"
    
    def __repr__(self) -> str:
        return self.value


import traceback
from typing import Any, Callable
from uuid import UUID, uuid4
from .logger import log


Event = str
EventCallbackFunction = Callable[[Any], None]
EventCallbackRemovalFunction = Callable[[None], None]
EventObserverList = dict[Event, dict[UUID, EventCallbackFunction]]


# Dispatcher Class
# Used to dispatch events in the application runtime
# Observers can listen to events and register a callback handler
# Called each time the event is triggered
class Dispatcher(object):
    def __init__(self) -> None:
        # Observers list
        self._observers: EventObserverList = {}
        
    def get_observers(self) -> EventObserverList:
        return self._observers

    def get_event_observers(self, event: Event) -> list[EventCallbackFunction]:
        # Return the event's observers callback list if event exists
        if event not in self._observers.keys():
            return []
        return list(self._observers[event].values())
    
    def add_observer(self, event: Event, handler: EventCallbackFunction) -> EventCallbackRemovalFunction:
        # If event is not present in observer list, we create an empty list
        if event not in self._observers:
            self._observers[event] = {}

        # Generate an uuid for the observer callback
        # Adding it to the list
        e_id = uuid4()
        self._observers[event][e_id] = handler

        # Returning a lambda function used to unregister obsrever
        return lambda : self.event_remove_observer(event, e_id)

    def event_remove_observer(self, event: Event, event_id: UUID) -> None:
        # If observer is present for this event with this id
        # We remove it from the list
        if event in self._observers:
            if event_id in self._observers[event]:
                del self._observers[event][event_id]

    def notify(self, event: Event, *args: Any, **kargs: Any) -> None:
        # We iterate the event's observers list
        # and run handler callcack for each
        if event in self._observers:
            for handler in self._observers[event].values():
                try:
                    if callable(handler):
                        handler(*args, **kargs)
                except:
                    # If an error occurs, we log it
                    # and continue the execution with the other handlers
                    log(f"Error while dispatching event {event}, trying the other handlers")
                    traceback.format_exc()                        

# Default dispatch for the application
# Users can create their own if needed
# Used in the UT for exemple
dispatcher: Dispatcher = Dispatcher()



from dataclasses import dataclass

from ..logger import log
from .payload import MessagePayload
from .enums import CommandOperationID, MessageModule, LogCommand, DBMCommand
from .utils import B642B,B2B64


@dataclass
class CommandGet:
    itemid: int
    
    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
        }

    def to_array(self) -> list:
        return [
            self.itemid,
        ]

@dataclass
class CommandSet:
    itemid: int
    
    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
        }

    def to_array(self) -> list:
        return [
            self.itemid,
        ]   
    

@dataclass
class CommandStartLogGetKpi:
    itemid: int
    index: int

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "index": self.index
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.index
        ]


@dataclass
class CommandStartDbmExportRow:
    itemid: int
    handle: int
    key: str
    index: int

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "handle": self.handle,
            "key": self.key,
            "index": self.index
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.handle,
            self.key,
            self.index
        ]


@dataclass
class CommandStartDbmCreateTable:
    itemid: int
    table: int
    nbrows: str
    rowsize: int

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "table": self.table,
            "nbrows": self.nbrows,
            "rowsize": self.rowsize
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.table,
            self.nbrows,
            self.rowsize
        ]
    
@dataclass
class CommandStartDbmImportRow:
    itemid: int
    table: int
    key: str
    data: bytes

    def __init__(self, itemid, table, key, data) -> None:
        self.itemid = itemid
        self.table = table
        self.key = key
        if(type(data) == str):
            data = B642B(data)
        self.data = bytearray(data)

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "table": self.table,
            "key": self.key,
            "data": B2B64(self.data),
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.table,
            self.key,
            self.data,
        ]


OP_TO_COMMAND_TYPES = {
    MessageModule.LOG.value: {
        CommandOperationID.CMD_GET.value : CommandGet,            
        CommandOperationID.CMD_START.value : {
            LogCommand.LOG_GET_KPI.value: CommandStartLogGetKpi
        }
    },

    MessageModule.DBM.value: {
        CommandOperationID.CMD_GET.value : CommandGet,            
        CommandOperationID.CMD_START.value : {
            DBMCommand.DBM_EXPORT_ROW.value: CommandStartDbmExportRow, 
            DBMCommand.DBM_CREATE_TABLE.value: CommandStartDbmCreateTable,
            DBMCommand.DBM_IMPORT_ROW.value: CommandStartDbmImportRow
        }
    },

    MessageModule.RUN.value: {
        CommandOperationID.CMD_GET.value : CommandGet
    },
    
    MessageModule.MAL.value: {
        CommandOperationID.CMD_GET.value : CommandGet
    }
    
}


def get_command_class(mod: int, op_id: int,  raw_data: list):
    if not mod in OP_TO_COMMAND_TYPES:
        return None

    if not op_id in OP_TO_COMMAND_TYPES[mod]:
        return None

    if op_id in (CommandOperationID.CMD_SET.value, CommandOperationID.CMD_START.value):
        if isinstance(raw_data,list):
            return OP_TO_COMMAND_TYPES[mod][op_id][raw_data[0]]
        else:
            return OP_TO_COMMAND_TYPES[mod][op_id].get(raw_data["data"]["itemid"], None)

    return OP_TO_COMMAND_TYPES[mod][op_id]


class CommandMessagePayload(MessagePayload):
    @classmethod
    def from_array(cls, mod: int, operation: int, raw_data: list) -> "MessagePayload":
        command_class = get_command_class(mod,operation,raw_data)
        if command_class is None:
            log(f"Operation {operation} unknown", logger_level="ERROR")
            return

        payload = cls()
        payload.command = command_class(*raw_data)

        return payload

    @classmethod
    def from_json(cls, mod: int, operation: int, raw_data: list) -> "MessagePayload":
        command_class = get_command_class(mod,operation,raw_data)
        if command_class is None:
            log(f"Operation {operation} unknown", logger_level="ERROR")
            return
            
        payload = cls()
        payload.command = command_class(**raw_data["data"])

        return payload

    def to_dict(self) -> dict:
        return {
            "data": self.command.to_dict()
        }



import msgpack
from dataclasses import dataclass
from io import BytesIO

from ..logger import log

from .payload import MessagePayload
from .enums import BLDCommandParameter, CommandOperationID, MALCommandParameter, MessageModule, LogCommand, DBMCommand, LogCommandParameter,\
                    DBMCommandParameter, RUNCommandParameter
from .process_dbm import DbmProcesser
from .utils import B2B64, B642B


@dataclass
class ResponseGetLogNbKpi:
    itemid: int    
    number: int

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "number": self.number
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.number
        ]
    

@dataclass
class ResponseGetMalLifeCycle:
    itemid: int    
    lifecycle: int

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "lifecycle": self.lifecycle
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.lifecycle
        ]


@dataclass
class ResponseSet:
    itemid: int    

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
        }

    def to_array(self) -> list:
        return [
            self.itemid,
        ]


@dataclass
class ResponseBldSetInitDone(ResponseSet):
    pass


@dataclass
class ResponseLogGetKpi:
    itemid: int
    execid: int
    result: int
    appuid: int
    typ: int
    kpid: int
    description: str
    value: int

    def to_dict(self) -> str:
        # Translate the kpid
        vmid,modelid = self.translate_kpi()
        return {
            "itemid": self.itemid,
            "execid": self.execid,
            "result": self.result,
            "appuid": self.appuid,
            "typ": self.typ,
            "kpid": self.kpid,
            "vmid": vmid,
            "modelid": modelid,
            "description": self.description,
            "value": self.value,
        }

    def to_array(self) -> list:
        vmid,modelid = self.translate_kpi()
        return [
            self.itemid,
            self.execid,
            self.result,
            self.appuid,
            self.typ,
            self.kpid,
            vmid,
            modelid,
            self.description,
            self.value
        ]

    def translate_kpi(self):
        modelid = 0
        vmid = 0
        kpid = self.kpid
        if kpid :
            mask = 2**8-1
            mask = mask<<24
            iscustom = (kpid & mask)>> 24

            if iscustom == 0:
                mask = 2**8-1
                mask = mask<<16
                vmid = (kpid & mask)>> 16

                mask = 2**8-1
                mask = mask<<8
                modelid = (kpid & mask)>> 8

                mask = 2**8-1
                typeid = (kpid & mask)
        return vmid,modelid


@dataclass
class ResponseDbmTableInfo:
    handle      : int
    typ         : int
    nbrows      : int
    rowsize     : int
    count       : int
    reserved    : int
    used        : int

    def to_dict(self) -> str:
        return {
            "handle"    : self.handle,
            "typ"       : self.typ,
            "nbrows"    : self.nbrows,
            "rowsize"   : self.rowsize,
            "count"     : self.count,
            "reserved"  : self.reserved,
            "used"      : self.used
        }

    def to_array(self) -> list:
        return [
            self.handle,
            self.typ,
            self.nbrows,
            self.rowsize,
            self.count,
            self.reserved,
            self.used
        ]


@dataclass(init=False)
class ResponseDbmGetInfo:    
    itemid: int    
    tables: list[ResponseDbmTableInfo]

    def __init__(self,itemid,tables):
        self.itemid = itemid
        self.tables = [ResponseDbmTableInfo(*t) for t in tables]


    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "tables": [t.to_dict() for t in self.tables]
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            [t.to_array for t in self.tables]
        ]


@dataclass
class ResponseDbmExportRow:
    itemid: int
    execid: int
    result: int
    attributes: dict

    def to_dict(self) -> str:
        att = DbmProcesser.from_msgpack(self.attributes)
        if att is None:
            att_as_dict = {}
        else:
            att_as_dict= att.to_dict()
        return {
            "itemid": self.itemid,
            "execid": self.execid,
            "result": self.result,
            "attributes": att_as_dict
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.execid,
            self.result,
            self.attributes
        ]


@dataclass
class ResponseRunGetOps:
    itemid: int
    ops: list

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "ops": self.ops
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.ops
        ]

@dataclass
class ResponseRunApp:
    uuid: bytes
    desc: str
    threads: int
    mem: list[int]
    ops: list[int]
    meta: dict

    def to_dict(self) -> str:

        stream = BytesIO(self.meta)
        unpacker = msgpack.Unpacker(stream, raw=False, strict_map_key=False)

        meta = unpacker.unpack()
        return {
            "uuid": B2B64(self.uuid),
            "desc": self.desc,
            "threads": self.threads,
            "mem": self.mem,
            "ops": self.ops,
            "meta": meta
        }
    
    def to_array(self) -> list:
        packer = msgpack.Packer()
        buffer = BytesIO()
        buffer.write(packer.pack(self.meta))
        buffer.seek(0)
        return [
            B642B(self.uuid),
            self.desc,
            self.threads,
            self.mem,
            self.ops,
            buffer.read()
        ]


@dataclass
class ResponseRunGetApps:
    itemid: int
    apps: list

    def to_dict(self) -> str:

        apps = [ResponseRunApp(*app).to_dict() for app in self.apps]

        return {
            "itemid": self.itemid,
            "apps": apps
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            [self.app.to_array() for self.app in self.apps]
        ]


@dataclass
class ResponseDbmCreateTable:
    itemid: int
    execid: int
    result: int

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "execid": self.execid,
            "result": self.result,
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.execid,
            self.result,
        ]


@dataclass
class ResponseDbmImportRow:
    itemid: int
    execid: int
    result: int

    def to_dict(self) -> str:
        return {
            "itemid": self.itemid,
            "execid": self.execid,
            "result": self.result,
        }

    def to_array(self) -> list:
        return [
            self.itemid,
            self.execid,
            self.result,
        ]


OP_TO_RESPONSE_TYPES = {
    MessageModule.LOG.value: {
        CommandOperationID.CMD_GET.value : {
            LogCommandParameter.LOG_NB_KPIS.value: ResponseGetLogNbKpi,
        },    

        CommandOperationID.CMD_START.value : {
            LogCommand.LOG_GET_KPI.value: ResponseLogGetKpi
        } 
    },

    MessageModule.DBM.value: {
        CommandOperationID.CMD_GET.value : {
            DBMCommandParameter.DBM_PARAM_INFO.value: ResponseDbmGetInfo,
        },    

        CommandOperationID.CMD_START.value : {
            DBMCommand.DBM_EXPORT_ROW.value: ResponseDbmExportRow,
            DBMCommand.DBM_CREATE_TABLE.value: ResponseDbmCreateTable,
            DBMCommand.DBM_IMPORT_ROW.value: ResponseDbmImportRow
        } 
    },

    MessageModule.RUN.value: {
        CommandOperationID.CMD_GET.value : {
            RUNCommandParameter.RUN_PARAM_OPS.value: ResponseRunGetOps,
            RUNCommandParameter.RUN_PARAM_APPS.value: ResponseRunGetApps,
        },
    },

    MessageModule.BLD.value: {
        CommandOperationID.CMD_SET.value : {
            BLDCommandParameter.BLD_PARAM_CORTEXM_INIT_DONE.value: ResponseBldSetInitDone,
        },
    },

    MessageModule.MAL.value: {
        CommandOperationID.CMD_GET.value : {
            MALCommandParameter.MAL_PARAM_LIFE_CYCLE.value: ResponseGetMalLifeCycle,
        },
    },

}


def get_response_class(mod: int, op_id: int, raw_data: list):
    if not mod in OP_TO_RESPONSE_TYPES:
        return None

    if not op_id in OP_TO_RESPONSE_TYPES[mod]:
        return None

    if op_id in (CommandOperationID.CMD_GET.value, CommandOperationID.CMD_SET.value, CommandOperationID.CMD_START.value):
        return OP_TO_RESPONSE_TYPES[mod][op_id].get(raw_data[0], None)

    return OP_TO_RESPONSE_TYPES[mod][op_id]


class ResponseMessagePayload(MessagePayload):
    @classmethod
    def from_array(cls, mod: int, operation: int, raw_data: list) -> "MessagePayload":
        response_class = get_response_class(mod, operation, raw_data)
        if response_class is None:
            log(f"Operation {operation} unknown", logger_level="ERROR")
            return

        payload = cls()
        payload.command = response_class(*raw_data)

        return payload

    @classmethod
    def from_json(cls, mod: int, operation: int, raw_data: list) -> "MessagePayload":
        response_class = get_response_class(mod, operation, raw_data)
        if response_class is None:
            log(f"Operation {operation} unknown", logger_level="ERROR")
            return
            
        payload = cls()
        payload.command = response_class(**raw_data["data"])

        return payload

    def to_dict(self) -> dict:
        return {
            "data": self.command.to_dict()
        }



import os
from uuid import UUID

from dataclasses import dataclass

from ..manager import Manager


ACT_SUB_TO_DEVICE = "sub-to-device"
ACT_UNSUB_FROM_DEVUCE = "unsub-from-device"
ACT_SEND_TO_DEVICE = "send-to-device"
ACT_GET_DEVICES_LIST = "get-device-list"
ACT_TRANSFERS_TO_WS_CLIENTS = "transfers-to-ws-clients"

EVT_ON_DEVICE_MSG = "on-device-msg"
EVT_ON_CLIENT_MSG = "on-client-msg"

VALID_ACTIONS = [
    ACT_GET_DEVICES_LIST,
    ACT_SEND_TO_DEVICE,
    ACT_UNSUB_FROM_DEVUCE,
    ACT_SUB_TO_DEVICE,
    ACT_TRANSFERS_TO_WS_CLIENTS
]


@dataclass
class WebsocketRequest:
    action: str
    device_name: str
    message: str
    token: str
    to_client_id: str
    from_client_id: str

    def __init__(self, action: str, device_name: str = None, message: str = None, token: str = None, to_client_id: str = None, from_client_id: str = None) -> None:
        self.action = action
        self.device_name = device_name
        self.message = message
        self.token = token
        self.to_client_id = str(to_client_id) if to_client_id is not None else None
        self.from_client_id = str(from_client_id) if to_client_id is not None else None

    def is_valid(self) -> bool:
        # If there is an WS AUTH TOKEN, we check that the request has the token
        auth_token = os.environ.get("GATEWAY_WS_AUTH_TOKEN", None)
        if auth_token and auth_token != self.token:
            return False

        # not valid if action is not in valid actions list
        if self.action not in VALID_ACTIONS:
            return False
        
        # if action is get device list, we don't need more arguments, so it's valid
        if self.action == ACT_GET_DEVICES_LIST:
            return True
        
        # Action is transfers msg to other ws clients so we need a message
        if self.action == ACT_TRANSFERS_TO_WS_CLIENTS and self.message:
            return True

        # if the device required is not registered it's not valid
        device = Manager.GetDevice(self.device_name)
        if not device:
            return False

        # if the action is send to device and there is no message, it's not valid
        if self.action == ACT_SEND_TO_DEVICE and not self.message: # TODO : Validate message
            return False
        
        return True

    def to_string(self) -> str:
        if not self.message:
            msg = "NULL"
        elif len(self.message) <= 350:
            msg = self.message
        else:
            msg = f"{self.message[:150]} ... {len(self.message)-300} ... {self.message[-150:]}"     
            
        return f"{self.action} - {self.device_name if self.device_name else 'NULL'} - {msg}"



import os
import argparse
import signal
from pathlib import Path

from .core.application import Application, CONFIG_DEFAULT_PATH
from .core.logger import log


CMD_NAME = "bondzai.gateway"
__version__ = "0.0.31"


def run():
    parser = argparse.ArgumentParser(
        prog=CMD_NAME,
        description="Davinsy Gateway"
    )

    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument(
        "-c", "--config",
        dest="config", help="Path to the YAML file for config.", 
        type=Path, default=CONFIG_DEFAULT_PATH
    )

    args = parser.parse_args()

    # TMP Fix for Windows CTRL + C not stopping threads
    if os.name == "nt":
        signal.signal(signal.SIGINT, signal.SIG_DFL)

    app = Application(Path(args.config))
        
    log(f"Bondzai Gateway {__version__}")

    app.run()



from websockets import WebSocketServerProtocol
from uuid import uuid4
import asyncio

from ..logger import log


class WebsocketClient(object):
    last_sent = 0
    
    def __init__(self, socket: WebSocketServerProtocol) -> None:
        self.socket = socket
        self.id = uuid4()

    async def _do_send(self, msg: str) -> None:
        await self.socket.send(msg)

    def send(self, msg: str) -> None:
        if len(msg) < 128:
            log(f"Sending message to client, msg = {msg}")
        else:
            log(f"Sending message to client, msg = {msg[:64]} ... {msg[-64:]}")

        new_loop = False
        try:
            loop = asyncio.get_event_loop()
            if not loop.is_running():
                # if the loop is not running we do as if there is none.
                raise
        except RuntimeError as e:
            if str(e).startswith('There is no current event loop in thread') or not loop.is_running():
                # If no loop or not running we create a new loop
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                new_loop = True
            else:
                log(f"Unable to send message to Websocket Client (no loop) : {str(e)}", logger_level="ERROR")
            
        if new_loop:
            # In the case of a new loop it means that we're not in
            # the websocket server thread. So we are in a connector thread
            # In that case we want to start asyncio and a new loop only to send
            # the message to the client then stop the loop
            loop.run_until_complete(self._do_send(msg))
            loop.close()
        else:
            # If it's not a new loop, then we're on the websocket server thread
            # And we just have to add a new task to the main running loop
            loop.create_task(self._do_send(msg))



import uuid
from .observer import dispatcher
from .events import DeviceEvents

TAG_AGENT_ID = 0x01
TAG_SESSION = 0x02

class Device(object):

    def __init__(self) -> None:
        self.out_data = []
        self.in_data = b""
        self.hid = 0
        self.ready = False # The device is not ready until it received the handshake
        self.name = None # The device name is set on handshake
        self.session = None # The session is set on handshake


    def set_name(self, hid: int, name: str) -> None:
        """ Called on handshake """
        
        if (self.ready):
            raise RuntimeError("Device is already ready")
        
        self.ready = True

        # set the hid and name
        self.hid = hid
        self.name = str(name, "utf-8")
    
        # generate a session id
        self.session = uuid.uuid4()

        # send the session id to the device
        self.out_data += [ 
            b"\x11" + 
            self.hid.to_bytes(1,"little") + 
            TAG_SESSION.to_bytes(1,"little") + 
            len(self.session.bytes).to_bytes(1,"little") + 
            self.session.bytes 
        ]

        # notify of connection
        dispatcher.notify(DeviceEvents.CONNECTED, self)


    def send(self, data: bytes) -> None:
        self.out_data += [b"\x13"+self.hid.to_bytes(1,"little")+data]



from uuid import UUID

from .logger import log
from .device import Device
from .websocket.client import WebsocketClient


class Manager(object):
    _devices: dict[str, Device] = {}
    _clients: dict[UUID, WebsocketClient] = {}
    _subscriptions: dict[str, dict[str, dict[UUID, WebsocketClient]]] = {}

    @classmethod
    def Init(cls) -> None:
        cls._devices = {}
        cls._subscriptions = {}

    @classmethod
    def RegisterDevice(cls, device: Device) -> None:
        cls._devices[str(device.name)] = device

    @classmethod
    def UnregisterDevice(cls, device: Device, remove_subs: bool = True) -> None:
        if device.name in cls._devices:
            del cls._devices[device.name]
        
        if remove_subs and device.name in cls._subscriptions:
            del cls._subscriptions[device.name]

    @classmethod
    def GetDevices(cls) -> list[Device]:
        return list(cls._devices.values())

    @classmethod
    def GetDevice(cls, device_name: str) -> Device:
        return cls._devices.get(device_name, None)

    @classmethod
    def RegisterClient(cls, client: WebsocketClient) -> None:
        cls._clients[str(client.id)] = client

    @classmethod
    def UnregisterClient(cls, client: WebsocketClient) -> None:
        if str(client.id) in cls._clients.keys():
            log(f"Removing client {client.id}")
            del cls._clients[str(client.id)]
            log(f"{cls._clients.keys()}")
        else:
            log(f"Unable to fin client {client.id} for removal")

    @classmethod
    def GetClients(cls) -> list[WebsocketClient]:
        return list(cls._clients.values())

    @classmethod
    def GetClient(cls, client_id: UUID) -> WebsocketClient:
        return cls._clients.get(client_id, None)

    # For now we only register to all events, but we have the posibility to 
    # register only for specific events

    @classmethod
    def SubToDeviceEvents(cls, device_name: str, client: WebsocketClient) -> None:
        if device_name not in cls._subscriptions:
            cls._subscriptions[device_name] = { "ALL": {} }
        cls._subscriptions[device_name]["ALL"][client.id] = client

    @classmethod
    def UnsubFromDeviceEvents(cls, device_name: str, client: WebsocketClient) -> None:
        if device_name in cls._subscriptions:
            if client.id in cls._subscriptions[device_name]["ALL"]:
                del cls._subscriptions[device_name]["ALL"][client.id]

    @classmethod
    def GetDeviceEventSubs(cls, device_name: str) -> list[WebsocketClient]:
        return list(cls._subscriptions.get(device_name, {}).get("ALL", {}).values())

    @classmethod
    def UnsubClient(cls, client: WebsocketClient) -> None:
        for device_name in cls._subscriptions:
            cls.UnsubFromDeviceEvents(device_name, client)


'''
 THIS IS A GENERATED FILE, DO NOT EDIT
 BONDZAI
          /"*._         _
      .-*'`    `*-.._.-'/
    < * ))     ,       ( 
      `*-._`._(__.--*"`.\
'''


from enum import Enum


class EventOperationID(Enum):
    EVT_EXT_EXCEPTION = 0x00
    EVT_EXT_DATA_IN = 0x01
    EVT_EXT_CMD_STATUS = 0x100
    EVT_EXT_LOG = 0x101
    EVT_EXT_VM_RESULT = 0x200
    EVT_EXT_ASL_RESULT = 0x201
    EVT_EXT_DATA_TRANSFORM_RESULT = 0x202
    EVT_EXT_DATA_ACQ_RESULT = 0x203
    EVT_EXT_ENROLL_RESULT = 0x204
    EVT_EXT_DATASET_PREPARATION_RESULT = 0x205
    EVT_EXT_TRAINING_RESULT = 0x206
    EVT_EXT_PROCESS_ACK = 0x210
    EVT_EXT_SET_MODE = 0x300
    EVT_EXT_CORRECTION = 0x301
    EVT_EXT_TRIGGER = 0x302
    EVT_EXT_SET_RECORD_MODE = 0x303
    EVT_EXT_LAUNCH_TRAIN = 0x304
    EVT_EXT_KILL = 0x305
    EVT_EXT_CUSTOM_1 = 0xF00
    EVT_EXT_CUSTOM_2 = 0xF01
    EVT_EXT_CUSTOM_3 = 0xF02
    EVT_EXT_CUSTOM_4 = 0xF03
    EVT_EXT_CUSTOM_5 = 0xF04
    EVT_EXT_CUSTOM_6 = 0xF05
    EVT_EXT_CUSTOM_7 = 0xF06
    EVT_EXT_CUSTOM_8 = 0xF07
    EVT_EXT_CUSTOM_9 = 0xF08
    EVT_EXT_CUSTOM_10 = 0xF09
    EVT_INT_EXCEPTION = 0x1000
    EVT_INT_DATA_STORED = 0x1001
    EVT_INT_DISCOVERY = 0x1002
    EVT_INT_TRANSPORT = 0x1003
    EVT_INT_OVERFLOW = 0x1004
    EVT_INT_DATA_UPDATED = 0x1100
    EVT_INT_READY_TO_PREPROC_BATCH = 0x1101
    EVT_INT_PREPROC_BATCH_DONE = 0x1102
    EVT_INT_TRIGGER_NEXT = 0x1103
    EVT_INT_FINAL = 0x1104
    EVT_INT_READY_TO_PREPROC_SINGLE = 0x1105
    EVT_INT_PREPROC_SINGLE_DONE = 0x1106
    EVT_INT_READY_TO_TRAIN = 0x1107
    EVT_INT_READY_TO_INFER = 0x1108
    EVT_INT_CORRECTION_DONE = 0x1109
    EVT_INT_INFER_DONE = 0x1200
    EVT_INT_TRAIN_DONE = 0x1201
    EVT_INT_POSTPROC_DONE = 0x1202
    EVT_INT_ITERATE_DB = 0x1300
    EVT_INT_ITERATE_DONE = 0x1301
    EVT_INT_CUSTOM_1 = 0x1F00
    EVT_INT_CUSTOM_2 = 0x1F01
    EVT_INT_CUSTOM_3 = 0x1F02
    EVT_INT_CUSTOM_4 = 0x1F03
    EVT_INT_YIELD_1 = 0x2000
    EVT_INT_YIELD_2 = 0x2001
    EVT_INT_YIELD_3 = 0x2002
    EVT_INT_YIELD_4 = 0x2003
    EVT_INT_FORCE_TRAIN = 0x3001
    EVT_INT_FORCE_DATA_IN = 0x3002
    EVT_INT_TIMER = 0x4000
    EVT_TRANSIENT_MMS_LOAD = 0xF001
    EVT_TRANSIENT_RAM_LOAD = 0xF002
    EVT_TRANSIENT_FLASH_LOAD = 0xF003
    EVT_MAX = 0xFFFF


class AgentAIMode(Enum):
    APP_AI_MODE_NOT_INIT = -1
    APP_AI_MODE_STANDBY = 0
    APP_AI_MODE_ENROLLEMENT = 1
    APP_AI_MODE_INFERENCE = 2


class AgentAIType(Enum):
    APP_AI_TYPE_NOT_INIT = -1
    APP_AI_TYPE_CLASSIFICATION = 0
    APP_AI_TYPE_REGRESSION = 1


class PPOperationUID(Enum):
    OPS_ID_EXT = 0x80000000
    OPS_ID = 0x00000000
    OP_CUST_MIN = OPS_ID_EXT
    OP_CUST_FIRST = 0x80000100
    OPS_MISC = 0x00000000
    OPS_MATH = 0x01000000
    OPS_DSP = 0x02000000
    OPS_IMG = 0x03000000
    OPS_STR = 0x04000000
    OPS_DBM = 0x05000000
    OPS_PST = 0x0E000000
    OPS_DIST = 0x0F000000
    OPS_UID_CP1D = 0x02
    OPS_UID_CP2D = 0x03
    OPS_UID_CP2X1D = 0x04
    OPS_UID_RMNFIRST1D_F32 = 0x05
    OPS_UID_LOGANF32 = 0x1000001
    OPS_UID_LOGA10F32 = 0x1000002
    OPS_UID_MEANF32 = 0x1000011
    OPS_UID_VARF32 = 0x1000012
    OPS_UID_RESHAPEF32 = 0x1000013
    OPS_UID_HISTOGRAMF32 = 0x1000014
    OPS_UID_VECTORMATRIX_PRODUCTF32 = 0x1000015
    OPS_UID_FLATTENMATRIXF32 = 0x1000016
    OPS_UID_TRANSPOSEMATRIXF32 = 0x1000017
    OPS_UID_SLICEF32 = 0x1000018
    OPS_UID_SLICEMATRIXROWSF32 = 0x1000019
    OPS_UID_SLICEMATRIXCOLSF32 = 0x100001A
    OPS_UID_IPSLICEF32 = 0x100001B
    OPS_UID_IPSLICEMATRIXROWSF32 = 0x100001C
    OPS_UID_IPSLICEMATRIXCOLSF32 = 0x100001D
    OPS_UID_ADD_VECTOR_BUFFERF32 = 0x1000030
    OPS_UID_FFT2048F32 = 0x2000011
    OPS_UID_FFT1024F32 = 0x2000012
    OPS_UID_FFT16F32 = 0x2000018
    OPS_UID_MFCC2048F32 = 0x2000019
    OPS_UID_MFCC1024F32 = 0x200001A
    OPS_UID_MEL2048F32 = 0x2000021
    OPS_UID_MEL1024F32 = 0x2000022
    OPS_UID_MEL16F32 = 0x2000028
    OPS_UID_DCT2048F32 = 0x2000031
    OPS_UID_DCT1024F32 = 0x2000032
    OPS_UID_DCT512F32 = 0x2000033
    OPS_UID_DCT256F32 = 0x2000034
    OPS_UID_DCT128F32 = 0x2000035
    OPS_UID_NOISEMIXF32 = 0x2000061
    OPS_UID_PWSPANF32 = 0x2000062
    OPS_UID_DCT2D = 0x3000001
    OPS_UID_RGB_TO_GRAYF32 = 0x3000002
    OPS_UID_PUSHSIG2DBM = 0x5000001
    OPS_UID_PST_BYPASS = 0xE000001
    OPS_UID_PST_EXT = 0x8E000001


class CommandOperationID(Enum):
    CMD_OPEN_SESSION = 0x00
    CMD_CLOSE_SESSION = 0x01
    CMD_SET = 0x100
    CMD_GET = 0x101
    CMD_DELETE = 0x102
    CMD_OPEN = 0x200
    CMD_CLOSE = 0x201
    CMD_WRITE = 0x202
    CMD_READ = 0x203
    CMD_START = 0x300
    CMD_STOP = 0x301
    CMD_SUBSCRIBE = 0x401
    CMD_UNSUBSCRIBE = 0x402


class LogCommandParameter(Enum):
    LOG_NB_KPIS = 6


class DBMCommandParameter(Enum):
    DBM_PARAM_INFO = 0x0000
    DBM_TABLE_ADDRESS_VM = 0x1000
    DBM_TABLE_ADDRESS_LBL = 0x1001
    DBM_TABLE_ADDRESS_DAT = 0x1002
    DBM_TABLE_ADDRESS_KEY = 0x1003
    DBM_TABLE_ADDRESS_CTX = 0x1004


class DBMAttributesVM(Enum):
    DBM_ATT_VM_DESC = 0x00
    DBM_ATT_VM_PREPROC_GRAPH_TRAIN = 0x01
    DBM_ATT_VM_PREPROC_GRAPH_INFER = 0x02
    DBM_ATT_VM_PREPROC_GRAPH_OTHER = 0x03
    DBM_ATT_VM_DEEPLOMATH = 0x04
    DBM_ATT_VM_POSTPROC = 0x05
    DBM_ATT_VM_PREPROC_GRAPH_FEATURE_EXTRACTION = 0x06
    DBM_ATT_VM_PREPROC_GRAPH_BACKGROUND_AUGMENTATION = 0x07
    DBM_ATT_VM_PREPROC_GRAPH_DATA_AUGMENTATION = 0x08
    DBM_ATT_VM_PREPROC_GRAPH_DATA_TRANSFORM = 0x09


class LogCommand(Enum):
    LOG_GET_KPI = 0x1100


class DBMCommand(Enum):
    DBM_EXPORT = 0x100
    DBM_EXPORT_ROW = 0x200
    DBM_IMPORT_ROW = 0x201
    DBM_DELETE_ROW = 0x202
    DBM_CHECK_POINT = 0x300
    DBM_RESTORE = 0x301
    DBM_DEL_CHECK_POINT = 0x302
    DBM_GET_CHECK_POINT = 0x303
    DBM_CREATE_TABLE = 0x400
    DBM_DELETE_TABLE = 0x401


class DBMTable(Enum):
    DBM_VM = 0x00
    DBM_LBL = 0x01
    DBM_DAT = 0x02
    DBM_KEY = 0x03
    DBM_CTX = 0x04
    DBM_SIG = 0x05
    DBM_GEN = 0x06
    DBM_TABLE_END = 0x07


class AgentTriggerType(Enum):
    TRIGGER_ON = 16
    TRIGGER_OFF = 17


class AgentRecordMode(Enum):
    DAVINSY_RECORD_OFF = 0
    DAVINSY_RECORD_BACKGROUND = 1
    DAVINSY_RECORD_SUBJECT = 2


class AgentFlowSteps(Enum):
    DATA_ACQUISITION = 1
    DATA_TRANSFORM = 2
    ENROLLEMENT = 3
    INFERENCE = 4
    DATASET_PREPARATION = 5
    TRAINING = 6


class MessageModule(Enum):
    BLD = 0x00
    COM = 0x01
    DBM = 0x02
    DLM = 0x03
    LOG = 0x04
    MAL = 0x05
    MMS = 0x06
    OTA = 0x07
    PRE = 0x08
    PST = 0x09
    RUN = 0x0A
    UTI = 0x0B
    UNK = 0x0C


class PPParam(Enum):
    OP_PARAM = 0x90000000
    OP_INOUT = 0xA0000000
    OP_INOUT_EMPTY = 0xA0000000
    OP_INOUT_MEMNEED = 0xA0000001
    OP_INOUT_VECT1DLEN = 0xA0000002
    OP_INOUT_VECT2DLEN = 0xA0000003
    OP_INOUT_VECT2X1DLEN = 0xA0000004
    OP_INOUT_VECT1D = 0xA0000011
    OP_INOUT_VECT2D = 0xA0000012
    OP_INOUT_VECT2X1D = 0xA0000013
    OP_PARAM_TYPE_PREPROC = 0x90001000
    OP_PARAM_TYPE_POSTPROC = 0x90002000
    OP_PARAM_TYPE_DIST = 0x90003000
    OP_PARAM_TYPE_RUN = 0x90004000
    OP_PARAM_TYPE_MISC = 0x90008000
    OP_PARAM_HPF = 0x90001001
    OP_PARAM_LPF = 0x90001002
    OP_PARAM_BPF = 0x90001003
    OP_PARAM_BSF = 0x90001004
    OP_PARAM_FFT = 0x90001101
    OP_PARAM_MFCC = 0x90001102
    OP_PARAM_MEL = 0x90001103
    OP_PARAM_DCT = 0x90001104
    OP_PARAM_DCT_2D = 0x90001105
    OP_PARAM_HIST = 0x90001106
    OP_PARAM_NOISEMIX = 0x90001301
    OP_PARAM_PWSPAN = 0x90001302
    OP_PARAM_COMPRESS = 0x90001402
    OP_PARAM_MATRIX = 0x90001403
    OP_PARAM_PRIVATE_TO_MATRIX = 0x90001410
    OP_PARAM_PRIVATE_TO_VECTOR = 0x90001411
    OP_PARAM_PRIVATE_TO_MATRIX_WITHBUFFERS = 0x90001412
    OP_PARAM_PRIVATE_TO_VECTOR_WITHBUFFERS = 0x90001413
    OP_PARAM_EMPTY = 0x90008000
    OP_PARAM_MODE = 0x90008001
    OP_PARAM_REMOVENFIRST = 0x90008003
    OP_PARAM_SHAPE = 0x90008005
    OP_PARAM_SLICE = 0x90008006
    OP_PARAM_WITHBUFFERS = 0x90008007


class KPITypes(Enum):
    KPI_FREE = 0
    KPI_TRAINING_TIME_START = 1
    KPI_TRAINING_TIME_STOP = 2
    KPI_TRAINING_DATASET_SIZE = 3
    KPI_TRAINING_NET_SIZE = 4
    KPI_TRAINING_WM_SIZE = 5
    KPI_INFER_TIME_START = 6
    KPI_INFER_TIME_STOP = 7
    KPI_INFER_WM_SIZE = 8
    KPI_PREPROC_TIME_START = 9
    KPI_PREPROC_TIME_STOP = 10
    KPI_PREPROC_I_TIME_START = 11
    KPI_PREPROC_I_TIME_STOP = 12
    KPI_POSTPROC_TIME_START = 13
    KPI_POSTPROC_TIME_STOP = 14
    KPI_OTHER_TIME_START = 15
    KPI_OTHER_TIME_STOP = 16
    KPI_SIZE_BYTE = 17
    KPI_SIZE_KBYTE = 18
    KPI_QI = 19
    KPI_OTHER_FLOAT = 20
    KPI_TRAINING_NET_LAYERS = 21


class RUNCommandParameter(Enum):
    RUN_PARAM_BREAK = 0x0100
    RUN_PARAM_APPS = 0x0200
    RUN_PARAM_OPS = 0x0300
    RUN_PARAM_EVQ_TIMEOUT = 0x0400


class MALCommandParameter(Enum):
    MAL_PARAM_NB_THREADS = 0x0100
    MAL_PARAM_DATA_SOURCES = 0x0200
    MAL_PARAM_LIFE_CYCLE = 0x0300
    MAL_PARAM_NVM_FILE = 0x8100
    MAL_PARAM_CLEAR_NVM = 0x8200
    MAL_PARAM_MONITORING = 0x2000


class BLDCommandParameter(Enum):
    BLD_PARAM_LINUX_LIB_NAMES = 0x8000
    BLD_PARAM_LINUX_PYTHON_SCRIPT_NAME = 0x8001
    BLD_PARAM_CORTEXM_INIT_DONE = 0x8002
    BLD_PARAM_AG_ID = 0x8003
    BLD_PARAM_LICENCE_STATUS = 0x8004


class ErrorCode(Enum):
    COM_ERR_NUMBER_BAD_SOURCE = 0x01000001
    MEM_WARN_BUFFER_OVERFLOW = 0x02000001
    PROCESS_WARN_TRAIN_EXCEPTION = 0x03000001
    PROCESS_WARN_TRAIN_EMPTY_DATASET = 0x03000002





class DeviceConnector(object):
    def open(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def run(self):
        raise NotImplementedError()



from gateway.core.observer import Dispatcher
from unittest.mock import Mock, call


def test_add_observers():
    d = Dispatcher()

    event1 = "event1"
    event2 = "event2"

    handler = lambda : None
    handler2 = lambda : None

    # Adding observers for events
    assert d.add_observer(event1, handler)
    assert d.add_observer(event2, handler)
    assert d.add_observer(event2, handler2)

    # Checking global observers counts
    obs_list = list(d.get_observers().keys())
    assert len(obs_list) == 2
    assert event1 in obs_list
    assert event2 in obs_list

    # Checking event1 observers
    e1_obs = d.get_event_observers(event1)
    assert len(e1_obs) == 1
    assert e1_obs[0] == handler

    # Checking event2 observers
    e2_obs = d.get_event_observers(event2)
    assert len(e2_obs) == 2
    assert e2_obs[0] in [handler, handler2]
    assert e2_obs[1] in [handler, handler2]

    # Checking that dispatcher get empty observers for unknown event
    assert len(d.get_event_observers("UNKNOWN-EVENT")) == 0


def test_remove_observer():
    d = Dispatcher()

    event = "e"
    
    handler = lambda : None
    handler2 = lambda : None

    # Adding two observers
    d.add_observer(event, handler)
    remove_obs = d.add_observer(event, handler2)

    # Checking that observers count is 2
    assert len(list(d.get_event_observers(event))) == 2
    # Removing observer 2
    remove_obs()
    # Checking that observers count is 1
    assert len(list(d.get_event_observers(event))) == 1


def test_notify_event():
    d = Dispatcher()
    event = "e"
    handler = Mock()

    # Adding event observer
    remove_handler = d.add_observer(event, handler)

    # Notifying event and checking that observer as been called
    d.notify(event)
    handler.assert_has_calls([call()])

    arg1 = 12
    arg2 = "test"
    
    # Notifying event with arguments and checking that observer as been called
    # in correct order with correct arguments
    d.notify(event, arg1, arg2)
    handler.assert_has_calls([
        call(),
        call(arg1, arg2)
    ])

    # Removing handler
    remove_handler()

    # Notifying event and checking that call stack hasn't moved
    d.notify(event)
    handler.assert_has_calls([
        call(),
        call(arg1, arg2)
    ])



import base64
import json
from typing import Any
import struct


def msg_serialize(o: Any) -> str:
    return json.dumps(o)


def msg_deserialize(o: str) -> Any:
    return json.loads(o)


def B2B64(data: bytes) -> str:
    return base64.b64encode(data).decode("utf-8")


def B642B(data: str) -> bytes:
    return base64.b64decode(data)


def unpack_buffer_to_list(buffer, is_little, data_len, pack_unit_format) -> list[Any]:
    buffer = bytearray(buffer)
    buffer_len = len(buffer)
    nb_items = int(buffer_len / data_len)
    struct_format = "".join([pack_unit_format for _ in range(0, nb_items)])
    if buffer_len != nb_items*data_len: # should not happened
        buffer_len = nb_items*data_len
        buffer = buffer[:buffer_len]
    return list(struct.unpack(f"{'<' if is_little else'>'}{struct_format}", buffer))


def pack_list_to_buffer(buffer: list[Any], is_little: bool, pack_unit_format: str) -> bytes:
    struct_format = "".join([pack_unit_format for _ in range(0, len(buffer))])
    return struct.pack(f"{'<' if is_little else' >'}{struct_format}", *buffer)


def get_step_uint32(**kwargs): # copy of generate_id in Dexim
    step_as_uint32 = int(b"00000000000000000000000000000000", 2) 
    
    #Reserved [4 bits]
    bits_to_shift = 4
    
    #Output [4 bits]
    output = kwargs.get("output", 0) 
    if output > (2**4-1) or output < 0:
        raise Exception("Invalid output")
    step_as_uint32 = step_as_uint32 | (output<<bits_to_shift)
    bits_to_shift +=4
    
    #Record mode (BACKGROUND/SUBJECT) [4 bits]
    record_mode = kwargs.get("record_mode", 0)
    if record_mode > (2**4-1) or record_mode < 0:
        raise Exception("Invalid record_mode")
    step_as_uint32 = step_as_uint32 | (record_mode<<bits_to_shift)
    bits_to_shift +=4

    #AI mode (ENROLL/INFER) [4 bits]
    ai_mode = kwargs.get("ai_mode", 0)
    if ai_mode > (2**4-1) or ai_mode < 0:
        raise Exception("Invalid ai_mode")
    step_as_uint32 = step_as_uint32 | (ai_mode<<bits_to_shift)
    bits_to_shift +=4
    
    #"Process" (DATA-TRANSFORM, CORRECTION, TRAIN…) [8 bits]
    process_id = kwargs.get("process_id", 0)
    if process_id > (2**8-1) or process_id < 0:
        raise Exception("Invalid process_id")
    step_as_uint32 = step_as_uint32 | (process_id<<bits_to_shift)
    bits_to_shift +=8

    #Source [7 bits]
    source_id = kwargs.get("source_id", -1)
    if source_id > (2**7-1) or source_id < 0:
        raise Exception("Invalid source_id")
    step_as_uint32 = step_as_uint32 | (source_id<<bits_to_shift)
    
    return step_as_uint32


def get_step_struct(id_as_uint32):
    standard_bit = id_as_uint32 >> 31
    if standard_bit != 0 :
        result = {
            "source_id":0,
            "process_id":0,
            "ai_mode":0,
            "record_mode":0,
            "output":0
        }
        return result
    #Reserved [4 bits]
    bits_to_shift = 4
    
    #Output [4 bits] 
    mask = 2**4-1
    mask = mask<<bits_to_shift
    output = (id_as_uint32 & mask)>> bits_to_shift
    bits_to_shift +=4
    
    #Record mode (BACKGROUND/SUBJECT) [4 bits]
    mask = 2**4-1
    mask = mask<<bits_to_shift
    record_mode = (id_as_uint32 & mask)>> bits_to_shift
    bits_to_shift +=4

    #AI mode (ENROLL/INFER) [4 bits]
    mask = 2**4-1
    mask = mask<<bits_to_shift
    ai_mode = (id_as_uint32 & mask)>> bits_to_shift
    bits_to_shift +=4
    
    #"Process" (DATA-TRANSFORM, CORRECTION, TRAIN…) [8 bits]
    mask = 2**8-1
    mask = mask<<bits_to_shift
    process_id = (id_as_uint32 & mask)>> bits_to_shift
    bits_to_shift +=8

    #Source [7 bits]
    mask = 2**7-1
    mask = mask<<bits_to_shift
    source_id = (id_as_uint32 & mask)>> bits_to_shift
    bits_to_shift +=7

    
    result = {
        "source_id":source_id,
        "process_id":process_id,
        "ai_mode":ai_mode,
        "record_mode":record_mode,
        "output":output
    }

    return result